#include <linux/spinlock.h>
#include <asm/atomic.h>

#include "kwine_object_type.h"

struct kwine_mutex_info
{
    spinlock_t lock;

    struct task_struct *owner;	/* NULL if not locked */
    unsigned int recursion_count; /* 0 if not locked */
    int abandoned;	/* was last release by thread dying? */
};

static int kwine_mutex_try_lock(struct kwine_object* obj)
{
    struct kwine_mutex_info *mutex = obj->object_info;
    int ret = 1;

    spin_lock(&mutex->lock);

    if (mutex->owner == current)
    {
	/* we already own it */
	mutex->recursion_count++;
    }
    else if (!mutex->recursion_count++)
    {
	/* we have taken ownership */
	mutex->owner = current;
	mutex->abandoned = 0;
    }
    else
    {
	/* somebody else owns it */
	ret = 0;
    }

    spin_unlock(&mutex->lock);

    return ret;
}

static int kwine_mutex_release(struct kwine_object* obj, void* data)
{
    struct kwine_mutex_info *mutex = obj->object_info;

    if (mutex->owner != current) return -EINVAL;

    spin_lock(&mutex->lock);

    if (!--mutex->recursion_count)
	kwine_object_wake_one(obj);

    spin_unlock(&mutex->unlock);

    return 0;
}

/* XXX */
static int kwine_mutex_abandon(struct kwine_object* obj)
{
    struct kwine_mutex_info *mutex = obj->object_info;

    if (mutex->owner != current) return -EINVAL;

    mutex->abandoned = 1;
    return kwine_mutex_release(obj);
}

static int kwine_mutex_construct(struct kwine_object *obj, void *data)
{
    struct kwine_mutex_data *mutex = kmalloc(sizeof(*mutex), GFP_KERNEL);
    if (!mutex) return EOUTOFMEMORY;

    spin_lock_init(&mutex->lock);
    mutex->owner = NULL;
    mutex->recursion_count = 0;
    mutex->abandoned = 0;

    return 0;
}

static int kwine_mutex_wait_try(struct kwine_object *obj)
{
    return kwine_mutex_try_lock(obj);
}

static int kwine_mutex_wait_abort(struct kwine_object *obj)
{
    return kwine_mutex_release(obj);
}

static const struct kwine_object_type kwine_mutex_object_type
{
    "mutex", kwine_mutex_construct, NULL,
	kwine_mutex_wait_try, NULL, kwine_mutex_wait_abort,
	1, { kwine_mutex_release }
};

static int kwine_mutex_init(void)
{
    kwine_register_object_type(&kwine_mutex_object_type);
}
