static unsigned int kwine_object_file_poll(struct file *file,
					   struct poll_table_struct *poll)
{
    struct inode *inode = file->f_dentry->d_inode;
    struct kwine_object *obj = inode->u.generic_ip;

    poll_wait(file, &obj->wait, poll);

    if (!obj->wait_try()) return 0;
    return obj->wait_commit();
}

static int kwine_object_file_ioctl(struct inode *inode, struct file *file,
				   unsigned int request, unsigned long arg)
{
    struct kwine_object *obj = inode->u.generic_ip;

    if (request >= obj->type->num_user_functions)
	return -ENOSYS;

    return obj->type->user_functions[request](obj, (void *)arg);
}

static const struct file_operations kwine_object_file_ops =
{
    poll: kwine_object_file_poll,
    ioctl: kwine_object_file_ioctl,
    /* flush: called from close */
    /* release: called from fput (when file count goes to 0) */
};

/*************************************************************************/

static struct dentry *kwine_get_directory_for_type(struct super_block *sb,
						   struct kwine_object_type *type)
{
    /* It would be nice if we stashed the dentry in the kwine_object_type
     * or something. */
    return d_lookup(sb->s_root,
		    &(const struct qstr){ type->name, strlen(type->name), 0 });
}

static int kwine_bind_object(struct super_block *sb,
			     struct kwine_object_type *type,
			     const char *name,
			     struct inode *inode)
{
    struct dentry *directory;
    struct dentry *dentry;

    directory = kwine_get_directory_for_type(sb, type);

    dentry = d_alloc(directory, &(cosnt struct qstr){ name, strlen(name), 0 });
    if (!dentry)
    {
	dput(directory);
	return -ENOMEM;
    }

    d_add(dentry, inode);

    return 0;
}

static int kwine_control_create_object(struct super_block *sb,
				       struct kwine_object_type *type,
				       const char *name)
{
    struct kwine_object *obj;
    struct inode *inode;
    int ret;

    obj = kwine_create_object(type);

    if (IS_ERR(obj)) return PTR_ERR(obj);

    inode = kwine_object_file_create(sb, obj);
    if (IS_ERR(inode))
    {
	kwine_delete_object(obj);
	return PTR_ERR(inode);
    }

    if (name != NULL)
    {
	ret = kwine_bind_object(type, name, inode);
	if (ret < 0) goto out_iput;
    }

    ret = kwine_install_fd(inode);
    if (ret < 0) goto out_unbind;

    return ret;

out_unbind:
    kwine_unbind_object(type, name);

out_iput:
    iput(inode);

    return ret;
}

static int kwine_validate_object_name(const char *name)
{
    if (name == NULL) return 1;
    if (name[0] == '\0') return 0;

    for (; *name != '\0'; name++)
	if (*name == '/') return 0;

    return 1;
}

static char *kwine_get_object_name(const char *user_name)
{
    char *name;

    if (user_name == NULL) return NULL;

    name = kmalloc(KWINE_MAX_PATH, GFP_KERNEL);
    if (!name) return ERR_PTR(-ENOMEM);

    if (strncpy_from_user(name, create_object.name, KWINE_MAX_PATH) == -EFAULT)
    {
	kfree(name);
	return ERR_PTR(-EFAULT);
    }

    if (!kwine_validate_object_name(name))
    {
	kfree(name);
	return ERR_PTR(-EINVAL);
    }

    return name;
}

static int kwine_control_file_ioctl(struct inode *inode, struct file *file,
				    unsigned int request, unsigned long arg)
{
    struct kwine_controller *ctl = inode->u.generic_ip;
    void *argp = (void *)arg;

    switch (request)
    {
    case KWINE_CREATE_OBJECT:
    {
	char *name = NULL;
	struct kwine_create_object create_object;

	if (copy_from_user(&create_object, argp, sizeof(create_object)))
	    return -EFAULT;

	name = kwine_get_object_name(create_object.name);
	if (IS_ERR(name)) return PTR_ERR(name);

	create_object.fd = kwine_control_create_object(inode->i_sb,
						       ctl->type, name);
	if (create_object.fd >= 0)
	{
	    if (copy_to_user(argp, &create_object, sizeof(create_object)))
	    {
		
	    }
	}

	return 0;
    }

    default:
	return -EINVAL;
    }
}

static void kwine_control_delete_inode(struct inode *inode)
{
    kfree(inode->u.generic_ip);
}

static const struct file_operations kwine_control_file_ops =
{
    ioctl: kwine_control_file_ioctl
};

struct inode *kwine_create_control_inode(struct super_block *sb,
					 const struct kwine_object_type *type)
{
    struct inode *inode;
    struct kwine_controller *controller;

    inode = kwine_create_inode(sb);
    if (!inode) return NULL;

    controller = kmalloc(sizeof(struct kwine_controller), GFP_KERNEL);
    if (!controller)
    {
	iput(inode);
	return NULL;
    }

    controller->type = type;
    controller->delete_inode = kwine_control_delete_inode;

    inode->u.generic_ip = controller;
    inode->i_fop = &kwine_control_file_ops;

    return inode;
}
